#!/usr/bin/php
<?php

# Nagios plugin for the Fritz!Box DECT 200 smart-power devices (and other "Smart Home" devices by AVM)
#
# (c) 2014 by Frederic Krueger / fkrueger-dev-checkfboxsmarthome@holics.at
#
# Licensed under Creative Commons Attribution-ShareAlike 3.0 Unported
# There is no warranty of any kind, explicit or implied, for anything this software does or does not do.
# 
# Updates for this piece of software could be available under the following URL:
#   http://dev.techno.holics.at/check_fbox_smarthome/
#
# Based on / Inspired by two of the scripts found at: http://www.tdressler.net/ipsymcon/fritz_aha.html


$cfgname = "/etc/nagios/check_fbox_smarthome.cfg";

$DEBUG = (0==1);


#### DONT EDIT BELOW HERE (unless you know what you are doing) ####

$minsensortoggletime = 15;      # allow toggling every 15 seconds at the most
$justdidwebcmd = "";            # only used in do_fbox_command, to get a toggle/switchon/switchoff cmd to be reflected in the status we output

$config = array();

# program info
$PROG_NAME     = 'check_fbox_smarthome';
$PROG_VERSION  = '0.0.1';
$PROG_EMAIL    = 'fkrueger-dev-checkfboxsmarthome@holics.at';
$PROG_URL      = 'http://dev.techno.holics.at/check_fbox_smarthome/';

# includes for nagios
if (file_exists("./utils.php")) { include ("./utils.php"); }
if (file_exists("/usr/lib/nagios/plugins/utils.php")) { include ("/usr/lib/nagios/plugins/utils.php"); }
if (file_exists("/usr/lib64/nagios/plugins/utils.php")) { include ("/usr/lib64/nagios/plugins/utils.php"); }
if (file_exists("/srv/nagios/libexec/utils.php")) { include ("/srv/nagios/libexec/utils.php"); }

# load configuration, if exists
if (file_exists($cfgname))
{
  $statinfo = stat ($cfgname);
  if (sizeof($statinfo) > 8)
  {
    $fmode = sprintf("0%o", 0777 & $statinfo['mode']);
    $omode = substr($fmode, -1, 1);
    if (($omode & 4) == 4)
    {
      print "CRITICAL - Configuration file with unsafe permissions ($cfgname has perms $fmode)\n";
      print "\n";
      print "!!! Others can read your stored passwords in the configuration file $cfgname.\n";
      print "    Run `chmod o-rwx $cfgname` to fix this.\n";
      print "\n";
      print "Refusing to run until that one is fixed.\n";
      print "\n";
      exit ($STATE_CRITICAL);
    } # end if abominable others-readable is found on the config file
  } # end if stat worked

  # load config if we are still here.
  $config = parse_ini_file ($cfgname, true);
} # end if cfgname exists


$foundconfiguredfboxhost = (0==1);
if ((isset($config)) and (sizeof($config) > 0))
{
  reset ($config);
  while (list ($fboxname, $fboxinfo) = each ($config))
  {
    if ((isset($fboxname)) and ($fboxname != "") and (isset($fboxinfo['host'])) and ($fboxinfo['host'] != "") and (isset($fboxinfo['pw'])) and ($fboxinfo['pw'] != ""))
    {
      $foundconfiguredfboxhost = (0==0);
    }
  }
}


# if configfile doesn't exist, or the hostname supplied can't be found in the config:
if (! $foundconfiguredfboxhost)
{
  print "UNKNOWN - Configuration file doesn't exist, or the host supplied doesn't exist in the config.\n";
  print "\n";
  print "You need to set up the configuration in file $cfgname .\n";
  print "\n";
  print "##### EXAMPLE CONFIG #####\n";
  print ";; a comment for your fbox with name 'fboxname'
;[fboxname]
;host='ip.or.host.name'
;pw='passwordtouse'
;; minimum time between toggling a switch is 15 seconds (lowest time allowed: 5 seconds)
;mintoggletime='15'
;
;; another fbox named 'fboxname2' (so multiple fboxen are supported)
;[fboxname2]
;host='fboxname2.hostname'
;pw='anotherpasswordtouse'\n";
  print "#####/EXAMPLE CONFIG #####\n\n";
  exit ($STATE_UNKNOWN);
}



# helpers
function usage ($detailed = "")
{
  global $PROG_NAME, $PROG_VERSION, $PROG_EMAIL, $PROG_URL;

  print "\n$myself <whattodo> <fboxhosts> <sensornames>\n";
  print "\n";
  print "    <whattodo>     can be 'read', 'toggle', 'switchon', 'switchoff'.\n";
  print "    <fboxhosts>    can be '*' for all or a search pattern on your config's section-names\n";
  print "    <sensornames>  can be '*' for all or a search pattern\n";
  print "\n";

  if ($detailed == "detailed")
  {
    print "
### NAGIOS COMMANDS DEFINITION
define command{
  command_name  check_fbox_smarthome
  command_line  /usr/bin/php \$USER1\$/contrib/check_fbox_smarthome.php \$ARG1\$ \$ARG2\$ \$ARG3\$
}

### NAGIOS SERVICECHECK DEFINITION
define service{
  use                             local-service
  host_name                       yourhost
  service_description             Smart Home
  check_command                   check_fbox_smarthome!'read'!'*'!'*'

# if you wanna use the following, you can use the accompanying execute_command.php (make sure it's reasonably secure for your environment):
#  notes_url                       /execute_command.php?arg1=toggle&cmd=check_fbox_smarthome&arg2=.*&arg3=.*
# 
# also, you need to set up an nrpe-server providing something like this:
#command[check_fbox_smarttemp] = /usr/bin/php /usr/lib/nagios/plugins/contrib/check_fbox_smarthome.php \$ARG1\$ \$ARG2\$ \$ARG3\$
}

";

    print '### PNP4NAGIOS check_commands configuration check_smarttemp.cfg
<?php

# Template to combine all selected data sets (of varying usefulness) provided by the plugin into one single graph
$opt[1] = " --title \"check_fbox_smarthome graph for " . $this->MACRO[\'DISP_HOSTNAME\'] . \' / \' . $this->MACRO[\'DISP_SERVICEDESC\'] . "\" ";
$def[1] = "";

foreach ($this->DS as $key => $val)
{
  $unit = "";
  if (preg_match("/_status$/", $val[\'NAME\'])) { $unit = "0=off,1=on"; }

  $def[1] .= rrd::def     ("var$key", $val[\'RRDFILE\'], $val[\'DS\'], "AVERAGE");
  $def[1] .= rrd::line2   ("var$key", rrd::color($key, 80) , $val[\'NAME\'] .(($unit != "") ? " ($unit)" : ""));
  if (preg_match("/_status$/", $val[\'NAME\']))
    { $def[1] .= rrd::gprint  ("var$key", array("LAST"), "%1.0lf ".$val[\'UNIT\']); }
  elseif (preg_match("/_total$/", $val[\'NAME\']))
    { $def[1] .= rrd::gprint  ("var$key", array("LAST","AVERAGE"), "%11.6lf ".$val[\'UNIT\']); }
  else
    { $def[1] .= rrd::gprint  ("var$key", array("LAST","AVERAGE"), "%12.7lf ".$val[\'UNIT\']); }

}
$def[1] .= rrd::comment("\\r");
$def[1] .= rrd::comment("check_fbox_smarthome graph template\\r");
$def[1] .= rrd::comment("Command " . $val[\'TEMPLATE\'] . "\\r");

?>
  
  
';
  } # end if is detailed usage info

  print "$PROG_NAME v$PROG_VERSION is licensed under Creative Commons Attribution-ShareAlike 3.0 Unported License.
There is no warranty of any kind, explicit or implied, for anything this software does or does not do.

The main page for this plugin can be found at: $PROG_URL

(c) 2014 by Frederic Krueger / $PROG_EMAIL
  
";
  exit ($STATE_UNKNOWN);
} # end func usage


function dbgprint ($mode="", $title="", $message="")
{
  global $DEBUG;
  $s = "";

  if ($mode != "") { $s = $s . "($mode) "; }
  if ($title != "") { $s = $s . "\"$title\": "; }
  if ($message != "") { $s = $s . "$message"; }

  if ($DEBUG) { print "$s\n"; }
} # end func dbgprint


function sanitize_for_fname ($str = "")
{
  $str = preg_replace("/ /", "_", $str);
  $str = preg_replace("/[^a-zA-Z\-\+0-9]/", "", $str);
  return ($str);
} # end func sanitize_for_fname


function load_url ($requrl="", $cnt=5)
{
  $nothingfound = (0==0);
  $content = "";
  while (($nothingfound) and ($cnt-- > 0))
  {
    $content = chop(file_get_contents($requrl));
    if (isset($content)) { $nothingfound = (0==1); }
  }

  if (!isset($content))
  {
    dbgprint ("dbg", "load_url", "No answer for query '$requrl'");
    return (0==0);
  }
  dbgprint ("dbg", "load_url", "Answer gotten for query '$requrl': " .((strlen($content) > 100) ? "big content gotten (" .strlen($content). " bytes)" : "\"$content\""));
  return ($content);
} # end func load_url


function block_toggle_switch ($prefix="", $fboxname="", $sensorname="", $mintoggletime = 0)
{
  if ($mintoggletime < 5) { $mintoggletime = 5; }   # minimum HARD limit
  $rc = (0==0);
  $prefix = sanitize_for_fname ($prefix);
  $fboxname = sanitize_for_fname ($fboxname);
  $sensorname = sanitize_for_fname ($sensorname);

  $fname = "/tmp/${prefix}_${fboxname}_${sensorname}.tmp";
  if (! file_exists($fname))
  {
    $fp = @fopen ($fname, "w");
    @fwrite ($fp, time());
    @fclose($fp);
    $rc = (0==1);
  } # end if no tmpfile exists yet
  else
  {
    $fp = @fopen ($fname, "r");
    $timestamp = @fgets ($fp, 99);
    @fclose ($fp);
    dbgprint ("dbg", "block_toggle_switch", "prefix '$prefix', fboxname '$fboxname', sensorname '$sensorname' => fname '$fname': got timestamp '$timestamp' (ts-time=" .($timestamp-time()). ")");
    if (time() - $timestamp > $mintoggletime)   # toggling is allowed!
    {
      $fp = @fopen ($fname, "w");
      @fwrite ($fp, time());
      @fclose ($fp);
      $rc = (0==1);
    } # end if toggling ok
    else
    {
      $rc = (0==0);
    } # end if toggling not yet ok
  } # end if tmpfile existed
  return ($rc);
} # end func block_toggle_switch


function get_sid ($loginurl = "", $fboxpw = "")
{

  # get challengestring from login-responsepage
  $http_response = file_get_contents ($loginurl);

  if (preg_match("/<Challenge>(\w+)<\/Challenge>/i", $http_response, $res))
    { $challenge = $res[1]; }
  if (preg_match("/<SID>([\da-f]+)<\/SID>/i", $http_response, $res))
    { $sid = $res[1]; }

  if ((isset($sid)) and (preg_match("/^[0]+$/",$sid)) and (isset($challenge)))
  {
    # sid is null, got challenge
    $sid = "";
    # build password response
    $pass = $challenge ."-". $fboxpw;
    # UTF-16LE encoding as required
    $pass = mb_convert_encoding ($pass, "UTF-16LE");
    # md5hash on top
    $md5 = md5($pass);
    # final answer string
    $challenge_response = $challenge ."-". $md5;
    # send to box
    $url = $loginurl. "?response=" .$challenge_response;
    $http_response = file_get_contents($url);
    # check answer
    if (preg_match("/<SID>([\da-f]+)<\/SID>/i", $http_response, $res))
    {
      # got answer with sid
      $sid = $res[1];
      if (!preg_match("/^[0]+$/",$sid))  # .. that even looks valid!
      {
        return ($sid);
      }
    }
  }

  return (null);
} # end func get_sid



function get_actors ($url, $sid)
{
  $actors = array();
  $actorlist = load_url ($url ."?switchcmd=getswitchlist&sid=". $sid);

  $a = explode(',', $actorlist);
  foreach ($a as $actor)
  {
    if (preg_match ('/^(([a-fA-F0-9]{2}[:|\-]?){6})$/', $actor))  # AIN is in hexadecimal, six numbers.
    {
      array_push($actors, $actor);
    }
  }
  return ($actors);
}


function do_fbox_command ($fboxname="", $fboxhost="", $fboxpw="", $sensornames="", $cmd = "")
{
  global $myself, $perfdata, $minsensortoggletime;

  # api urls
  $loginurl  = "http://" .$fboxhost. "/login_sid.lua";
  $logouturl = "http://" .$fboxhost. "/home/home.lua?logout=1";
  $ahaurl    = "http://" .$fboxhost. "/webservices/homeautoswitch.lua";

  # login
  $sid = get_sid ($loginurl, $fboxpw);

  # set up logout url here
  $logouturl = $logouturl . "&sid=$sid";

  if (!isset($sid))
  {
    dbgprint ("dbg", "Fritz-Login", "Login failed");
    print "CRITICAL - Logon to Fbox on host '$fboxhost' (section $fboxname) failed\n";
    exit ($STATE_CRITICAL);
  }

  # get actor names via aha api
  $actors = get_actors ($ahaurl, $sid);
  
  if (sizeof($actors) < 0)
  {
    dbgprint ("dbg", "Fritz-GetActors", "No actors found for section $fboxname");
    load_url ($logouturl);
    print "UNKNOWN - No actors found for section $fboxname'\n";
    exit ($STATE_UNKNOWN);
  }
  
  # get actor data via provided ain above
  foreach ($actors as $ain)
  {
    $url = $ahaurl ."?sid=$sid&ain=$ain";

    # query device values
    $name = load_url ($url ."&switchcmd=getswitchname");

    if (!isset($name)) { continue; }
    if (!preg_match ("/$sensornames/", $name)) { continue; }
  
    # parse provided data
    $status    = load_url ($url ."&switchcmd=getswitchstate");
    $power     = load_url ($url ."&switchcmd=getswitchpower");
    $new_total = load_url ($url ."&switchcmd=getswitchenergy");
  
    if ( (!isset($name)) or (!isset($status)) or (!isset($power)) or (!isset($new_total)) )
    {
      dbgprint ("dbg", "get_data", "Got no reply for some of the fields in section $fboxname.. Skipping the rest.");
      continue;
    }
  
    # prepare output
    if (preg_match ("/^\d+$/", $power))
      { $actual = $power/1000; } # mW->W
    else
      { $actual = ""; }

    if (preg_match ("/^\d+$/", $new_total))
    {
      $new_total = $new_total/1000; # Wh ->KWh
      $energy = $new_total-$total;
    }
    else
      { $energy = ""; }
  
    if ( (empty($energy)) and ($energy !== "0") ) # "0" equals "" and !isset.. yeah right, i know why i prefer perl
    {
      dbgprint ("dbg", "get_data", "No usable data gotten for some of the fields in section $fboxname.");
      continue;
    }
  
    # store data for later output
    dbgprint ("dbg", "got_data", "$name: " .(($status == "1") ? "on" : "off"). " (Actual: $actual W, Diff: $energy Wh, Power: $power mW, Total: $new_total Kwh)");

    if (!isset($perfdata[$fboxname]))
      { $perfdata[$fboxname] = array(); }

    if (!isset($perfdata[$fboxname][$ain]))
    {
      $perfdata[$fboxname][$ain] = [ 'name' => $name, 'actual' => $actual, 'energy' => $energy, 'new_total' => $new_total, 'status' => $status ];
    }
    else
    {
      dbgprint ("dbg", "got_data", "Got duplicate entry for ain '$ain' in section $fboxname.. Skipping data.");
    }

    if (($cmd == "toggle") or ($cmd == "switchon") or ($cmd == "switchoff"))
    {
      $localmintoggletime = $config[$fboxname]['mintoggletime'];
      if ((!isset($localmintoggletime)) or ($localmintoggletime == "")) { $localmintoggletime = $minsensortoggletime; }
      if (! block_toggle_switch ($myself, $fboxname, $name, $localmintoggletime))
      {
        $webcmd = "";
        if ($cmd == "toggle")
          { $webcmd = (($perfdata[$fboxname][$ain]['status'] > 0) ? "setswitchoff" : "setswitchon"); }
        elseif ($cmd == "switchoff")
          { $webcmd = "setswitchoff"; }
        elseif ($cmd == "switchon")
          { $webcmd = "setswitchon"; }

        dbgprint ("dbg", "toggle", "$fboxname:$ain> doing $webcmd");
        load_url ($url ."&switchcmd=$webcmd");

        if ($webcmd == "setswitchon")
          { $justdidwebcmd = "switchon"; }
        elseif ($webcmd == "setswitchoff")
          { $justdidwebcmd = "switchoff"; }
      } # end if toggleswitch is not blocked any more
      else
      {
        print "UNKNOWN - Changing power state of $fboxname:$name not yet allowed.\n";
      }
    } # end if cmd == toggle

  } # end foreach ain found in actors list

  load_url ($logouturl);

  return ($rc);
} # end func do_fbox_command








### init
$ARGV = $_SERVER['argv'];
$myself = $ARGV[0];

# our vars
$cmd = "";
$fboxhosts = "";
$sensornames = "";

$perfdata = "";


### args
if ((isset($ARGV[1])) and (isset($ARGV[2])) and (isset($ARGV[3])))
{
  $cmd = "read";
  if ($ARGV[1] == "toggle") { $cmd = "toggle"; }
  if ($ARGV[1] == "switchon") { $cmd = "switchon"; }
  if ($ARGV[1] == "switchoff") { $cmd = "switchoff"; }
  $fboxhosts = ($ARGV[2] != "") ? $ARGV[2] : "*";
  $sensornames = ($ARGV[3] != "") ? trim($ARGV[3]) : "*";

  if ($fboxhosts == "*") { $fboxhosts = ".*"; }
  if ($sensornames == "*") { $sensornames = ".*"; }
}

dbgprint ("dbg", "args", "cmd: '$cmd', fboxhosts: '$fboxhosts', sensornames: '$sensornames'");


### print usage, if needed or wanted
if (($ARGV[1] == "-h") or ($ARGV[1] == "--help"))
  { usage("detailed"); }
elseif (($cmd == "") or ($fboxhosts == "") or ($sensornames == ""))
  { usage(); }

### parse through config sections, looking for the supplied hosts/sensors
reset($config);
while (list($fboxname, $info) = each($config))
{
  if ((isset($fboxname)) and ($fboxname != ""))
  {
    if (preg_match("/$fboxhosts/", $fboxname))    # found matching host
    {
      if ((isset($info['host'])) and ($info['host'] != "") and (isset($info['pw'])) and ($info['pw'] != ""))
      {
        dbgprint ("dbg", "do_fbox_command", "cmd '$cmd' for host '" .$info['host']. "' (section $fboxname), sensornames '$sensornames'");
        do_fbox_command ($fboxname, $info['host'], $info['pw'], $sensornames, $cmd);
      } # end if got validlooking data in current config entry
    }
  }
}


### now for the nagios output
$nagiosrc = $STATE_UNKNOWN;
$nagiosoutput = "";
$perfdatastring = "";

if (sizeof($perfdata) <= 0)
{
  $nagiosoutput = "UNKNOWN - no sensor data found on host $fboxhost";
  $nagiosrc = $STATE_UNKNOWN;
} # end if no perfdata gotten

else

{
  # create perfdata string
  reset ($perfdata);
  while (list($fboxname, $info) = each ($perfdata))
  {
    while (list($ain, $data) = each ($info))
    {
      $perfdatastring = $perfdatastring . " ${fboxname}_" .$data['name']. "_energy=" .($data['energy']/1000). "kW;;;; ${fboxname}_" .$data['name']. "_actual=" .($data['actual']/1000). "kW;;;; ${fboxname}_" .$data['name']. "_total=" .$data['new_total']. "kW/h;;;; ${fboxname}_" .$data['name']. "_status=" .$data['status']. ";;;; ";
    } # end parsing fboxhost-data
  } # end parsing fboxhosts


  # TODO add crit/warn values later; will need a full overhaul of how to check which value against which values
  $nagiosrc = $STATE_OK;

  # then the nagios output
  reset ($perfdata);
  while (list($fboxname, $info) = each ($perfdata))
  {
    while (list($ain, $data) = each ($info))
    {
      if ($nagiosoutput == "")  # first line
      {
        $justdidwebcmd = ($justdidwebcmd != "") ? $justdidwebcmd : (($data['status'] > 0) ? "switchon" : "switchoff");
        $nagiosoutput = $nagiosoutput . "OK - $fboxname:" .$data['name']. "> Sensor " .(($justdidwebcmd == "switchon") ? "is active and uses " .$data['actual']. "W power" : "is inactive at the moment.") ." |$perfdatastring\n";
      }
      else                      # other lines
      {
        $nagiosoutput = $nagiosoutput . "OK - $fboxname:" .$data['name']. "> Sensor " .(($data['status'] > 0) ? "is active and uses " .$data['actual']. "W power" : "is inactive at the moment.") ."\n";
      }
    } # end parsing fboxhost-data
  } # end parsing fboxhosts
} # end if got perfdata


### and output everything:
print "$nagiosoutput\n";    # includes perfdatastring on first line
exit ($nagiosrc);





?>
